import logging
from typing import Optional

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.types.enums import ProcessingType
from picsellia.types.schemas import ProcessingSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class Processing(Dao):
    def __init__(self, connexion: Connexion, data: dict):
        Dao.__init__(self, connexion, data)

    @property
    def name(self) -> str:
        """Name of this (Processing)"""
        return self._name

    @property
    def type(self) -> ProcessingType:
        """Type of this (Processing)"""
        return self._type

    @property
    def docker(self) -> str:
        """Docker image of this (Processing)"""
        return f"{self._docker_image}:{self._docker_tag}"

    def __str__(self):
        return f"{Colors.GREEN}Processing '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/processing/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> ProcessingSchema:
        schema = ProcessingSchema(**data)
        self._name = schema.name
        self._type = schema.type
        self._docker_image = schema.docker_image
        self._docker_tag = schema.docker_tag
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        docker_image: Optional[str] = None,
        docker_tag: Optional[str] = None,
        description: Optional[str] = None,
        default_parameters: Optional[dict] = None,
        default_cpu: Optional[int] = None,
        default_gpu: Optional[int] = None,
    ) -> None:
        """Update docker_image, description or default_parameters of (Processing).

        Examples:
            ```python
            processing.update(docker_image='new-image', docker_tag='1.2.0')
            ```

        Arguments:
            docker_image (str, optional): New docker image of this (Processing). Defaults to None.
            docker_tag (str, optional): New docker tag of this (Processing). Defaults to None.
            description (str, optional): New description of the (Processing). Defaults to None.
            default_parameters (dict, optional): New default parameters of the (Processing). Defaults to None.
            default_cpu (str or InferenceType, optional): New default cpu of the (Processing). Defaults to None.
            default_gpu (str or InferenceType, optional): New default gpu of the (Processing). Defaults to None.
        """
        payload = {
            "docker_image": docker_image,
            "docker_tag": docker_tag,
            "description": description,
            "default_parameters": default_parameters,
            "default_cpu": default_cpu,
            "default_gpu": default_gpu,
        }
        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"/sdk/processing/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this processing from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            processing.delete()
            ```
        """
        self.connexion.delete(f"/sdk/processing/{self.id}")
        logger.info(f"{self} deleted.")



import logging
from typing import Optional
from uuid import UUID

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.label import Label
from picsellia.types.schemas import ClassificationSchema

logger = logging.getLogger("picsellia")


class Classification(Dao):
    def __init__(self, connexion: Connexion, annotation_id: UUID, data: dict) -> None:
        Dao.__init__(self, connexion, data)
        self._annotation_id = annotation_id

    @property
    def annotation_id(self) -> UUID:
        """UUID of (Annotation) holding this (Classification)"""
        return self._annotation_id

    @property
    def label(self) -> Label:
        """(Label) of this (Classification)"""
        return self._label

    @property
    def text(self) -> Optional[str]:
        """OCR text of this (Classification)"""
        return self._text

    def __str__(self):
        return f"{Colors.BLUE}Classification with label {self.label.name} on annotation {self.annotation_id} {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/classification/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> ClassificationSchema:
        schema = ClassificationSchema(**data)
        self._label = Label(self.connexion, schema.label.model_dump())
        self._text = schema.text
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        label: Optional[Label] = None,
        text: Optional[str] = None,
    ) -> None:
        """Update this classification with another label.

        Examples:
            ```python
            classification.update(label=label_plane)
            ```

        Arguments:
            label: (Label) to update this (Classification). Defaults to None.
            text (str, optional): New ocr text of this (Classification). Defaults to None.
        """
        payload = {}
        if label is not None:
            payload["label_id"] = label.id
        if text is not None:
            payload["text"] = text
        assert payload != {}, "You can't update this point with no data to update"
        r = self.connexion.patch(
            f"/api/classification/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this classification from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            classification.delete()
            ```
        """
        self.connexion.delete(f"/api/classification/{self.id}")
        logger.info(f"{self} deleted.")



import abc
import logging
from datetime import date
from typing import ClassVar, Generic, Optional, TypeVar, Union
from uuid import UUID

from beartype import beartype
from orjson import orjson

from picsellia.decorators import exception_handler
from picsellia.sdk.campaign.abstract_step import AbstractCampaignStep
from picsellia.sdk.dao import Dao
from picsellia.sdk.job import Job
from picsellia.sdk.worker import Worker
from picsellia.types.enums import CampaignStepType, WorkerType
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")

TCampaignStep = TypeVar("TCampaignStep", bound=AbstractCampaignStep)


class AbstractCampaign(Dao, abc.ABC, Generic[TCampaignStep]):
    _base_path: ClassVar[str]
    _allowed_step_types: ClassVar[list[CampaignStepType]]
    _worker_type: ClassVar[WorkerType]

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"{self._base_path}/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete a campaign.

        :warning: **DANGER ZONE**: Be very careful here!

        It will remove this campaign from our database, all tasks and assignments will be removed.
        This will not delete annotations, reviews and assets.

        Examples:
            ```python
            foo_campaign.delete()
            ```
        """
        self.connexion.delete(f"{self._base_path}/{self.id}")
        logger.info(f"{self} deleted")

    def _update(
        self,
        description: Optional[str] = None,
        instructions_object_name: Optional[str] = None,
        instructions_text: Optional[str] = None,
        end_date: Optional[date] = None,
        auto_add_new_assets: Optional[bool] = None,
        auto_close_on_completion: Optional[bool] = None,
        default_entry_step: Optional[TCampaignStep] = None,
        annotated_entry_step: Optional[TCampaignStep] = None,
    ):
        if default_entry_step and default_entry_step.campaign_id != self.id:
            raise ValueError("Given default entry step must be in this campaign")

        if annotated_entry_step and annotated_entry_step.campaign_id != self.id:
            raise ValueError("Given annotated entry step must be in this campaign")

        payload = {
            "description": description,
            "instructions_object_name": instructions_object_name,
            "instructions_text": instructions_text,
            "end_date": end_date,
            "auto_add_new_assets": auto_add_new_assets,
            "auto_close_on_completion": auto_close_on_completion,
        }

        if default_entry_step:
            payload["default_entry_step_id"] = default_entry_step.id

        if annotated_entry_step:
            payload["annotated_entry_step_id"] = annotated_entry_step.id

        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"{self._base_path}/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def add_step(
        self,
        name: str,
        type: Union[CampaignStepType, str],
        description: Optional[str] = None,
        assignees: Optional[list[tuple[Union[Worker, UUID], Union[float, int]]]] = None,
        order: Optional[int] = None,
        sample_rate: Optional[float] = None,
    ) -> TCampaignStep:
        """Add a step on this campaign.

        Examples:
            ```python
            workers = foo_dataset.list_workers()
            # In first step, 2 over 3 annotation task will be assigned to worker 1, third one is going to worker 0
            foo_campaign.add_step(
                name="annotation-step",
                type="ANNOTATION",
                description="annotation step",
                assignees=[(workers[0], 1.0), (workers[1], 2.0)]
            )
            # In second step, all review task will be assigned to worker 2
            foo_campaign.add_step(
                name="review-step",
                type="REVIEW",
                description="review step",
                assignees=[(workers[2], 1)]
            )
            ```
        Arguments:
            name (str): name of the step
            type (str or CampaignStepType): Type of the step: can be ANNOTATION or REVIEW
            description (str, optional): Description of the step. Defaults to None.
            assignees (list of tuple of (worker or UUID), float, optional): Can be used to assign workers to this step.
                Defaults to None.
            order (int, optional): Index where to insert the step in the workflow, the step will be appended at the end
                if nothing is specified. Defaults to None.
            sample_rate (float, optional): sample rate of this step. Defaults to None.
        Returns:
            dict, data of this step
        """
        campaign_step_type = CampaignStepType.validate(type)
        if campaign_step_type not in self._allowed_step_types:
            raise ValueError(
                f"Campaign step type must be in {self._allowed_step_types}"
            )

        payload = {
            "name": name,
            "type": campaign_step_type,
            "description": description,
            "order": order,
            "sample_rate": sample_rate,
        }
        if assignees:
            workers = []
            for worker, weight in assignees:
                if isinstance(worker, Worker):
                    if worker.type != self._worker_type:
                        raise ValueError(
                            f"Assignee {worker} cannot be used for this campaign."
                        )
                    worker_id = worker.id
                else:
                    worker_id = worker
                workers.append({"worker_id": worker_id, "weight": weight})

            payload["assignees"] = workers

        r = self.connexion.post(
            f"{self._base_path}/{self.id}/steps",
            data=orjson.dumps(payload),
        ).json()
        logger.info(f"Step {name} has been added to this campaign")
        return self._build_step(r)

    @abc.abstractmethod
    def _build_step(self, data: dict):
        pass

    @exception_handler
    @beartype
    def list_steps(self) -> list[TCampaignStep]:
        """List all the steps of this campaign.

        Returns:
            a list of steps of this campaign as python dict
        """
        r = self.connexion.get(f"{self._base_path}/{self.id}/steps").json()
        return [self._build_step(item) for item in r["items"]]

    @exception_handler
    @beartype
    def _sync_assets(
        self, existing_annotations_step_id: Union[str, UUID, None] = None
    ) -> Job:
        r = self.connexion.get(
            f"{self._base_path}/{self.id}/assignments/sync/available"
        ).json()
        if "is_sync_available" in r and r["is_sync_available"] is False:
            reason = r["reason"]
            raise ValueError(f"Sync not available because {reason}")

        item_count = r["unsynced_item_count"]

        payload = {}
        if existing_annotations_step_id:
            payload["existing_annotations_step_id"] = existing_annotations_step_id

        r = self.connexion.post(
            f"{self._base_path}/{self.id}/assignments/sync",
            data=orjson.dumps(payload),
        ).json()
        self.refresh(r["campaign"])

        logger.info(
            f"Sync job is starting, {item_count} assets will be added to the campaign."
        )
        return Job(self.connexion, r["job"], version=2)



import logging
import os
from collections.abc import Iterable, Mapping
from pathlib import Path
from typing import IO, Any, Optional, Union
from uuid import UUID

import orjson
import requests
from beartype import beartype
from picsellia_connexion_services import TokenServiceConnexion
from requests import Session
from requests.exceptions import ConnectionError

import picsellia.exceptions as exceptions
from picsellia import __version__
from picsellia.decorators import exception_handler, retry
from picsellia.services.upload.file import FileUploader
from picsellia.types.enums import ObjectDataType
from picsellia.utils import handle_response

logger = logging.getLogger("picsellia")

DEFAULT_TIMEOUT = 30


class Connexion(TokenServiceConnexion):
    def __init__(
        self,
        host: str,
        api_token: str,
        content_type: str = "application/json",
        session: Optional[Session] = None,
    ) -> None:
        super().__init__(
            host, api_token, authorization_key="Bearer", content_type=content_type
        )
        self._connector_id = None
        self._organization_id = None
        self.add_header("User-Agent", f"Picsellia-SDK/{__version__}")
        if session is not None:
            self.session = session

    @property
    def connector_id(self):
        if self._connector_id is None:
            raise exceptions.NoConnectorFound(
                "This organization has no default connector, and connect retrieve and upload files."
            )
        return self._connector_id

    @connector_id.setter
    def connector_id(self, value):
        self._connector_id = value

    @property
    def organization_id(self):
        return self._organization_id

    @organization_id.setter
    def organization_id(self, value):
        self._organization_id = value
        self.add_header("X-Picsellia-Organization", str(self._organization_id))

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, Connexion):
            return self.host == __o.host and self.api_token == __o.api_token

        return False

    @handle_response
    @retry(requests.ConnectionError)
    def get(self, path: str, params: Optional[dict] = None, stream=False):
        return super().get(path=path, params=params, stream=stream)

    @handle_response
    @retry(requests.ConnectionError)
    def xget(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
        stream=False,
    ):
        return super().xget(path=path, data=data, params=params, stream=stream)

    @handle_response
    @retry(requests.ConnectionError)
    def post(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
        files: Optional[Any] = None,
    ):
        return super().post(path=path, data=data, params=params, files=files)

    @handle_response
    @retry(requests.ConnectionError)
    def put(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
    ):
        return super().put(path=path, data=data, params=params)

    @handle_response
    @retry(requests.ConnectionError)
    def patch(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
    ):
        return super().patch(path=path, data=data, params=params)

    @handle_response
    @retry(requests.ConnectionError)
    def delete(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
    ):
        return super().delete(path=path, data=data, params=params)

    ##############################################################
    # ------------------------- UPLOAD ------------------------- #
    ##############################################################
    @exception_handler
    @beartype
    def _generate_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        connector_id: Optional[UUID] = None,
        context: Optional[dict[str, UUID]] = None,
    ) -> str:
        if connector_id is None:
            connector_id = self.connector_id

        payload = {"filename": filename, "type": object_name_type}
        if context:
            payload["context"] = context

        r = self.post(
            path=f"/api/organization/{self.organization_id}/connector/{connector_id}/generate_object_name",
            data=orjson.dumps(payload),
        ).json()
        return r["object_name"]

    @beartype
    def generate_data_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in [
            ObjectDataType.DATA,
            ObjectDataType.DATA_PROJECTION,
        ]:
            raise RuntimeError(
                f"Cannot generate data object name with type {object_name_type}"
            )
        return self._generate_object_name(filename, object_name_type, connector_id)

    @beartype
    def generate_dataset_version_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        dataset_version_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in [ObjectDataType.CAMPAIGN_FILE]:
            raise RuntimeError(
                f"Cannot generate dataset version object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id,
            context={"dataset_version_id": dataset_version_id},
        )

    @beartype
    def generate_deployment_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        deployment_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in [ObjectDataType.REVIEW_CAMPAIGN_FILE]:
            raise RuntimeError(
                f"Cannot generate deployment object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id,
            context={"deployment_id": deployment_id},
        )

    @beartype
    def generate_job_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        job_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in [ObjectDataType.LOGGING]:
            raise RuntimeError(
                f"Cannot generate job object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"job_id": job_id},
        )

    @beartype
    def generate_experiment_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        experiment_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in (
            ObjectDataType.ARTIFACT,
            ObjectDataType.LOG_IMAGE,
            ObjectDataType.LOGGING,
        ):
            raise RuntimeError(
                f"Cannot generate experiment object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"experiment_id": experiment_id},
        )

    @beartype
    def generate_model_version_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        model_version_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in (
            ObjectDataType.MODEL_THUMB,
            ObjectDataType.MODEL_FILE,
        ):
            raise RuntimeError(
                f"Cannot generate model version object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"model_version_id": model_version_id},
        )

    @beartype
    def generate_report_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        dataset_version_id: UUID,
        report_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type != ObjectDataType.AGENTS_REPORT:
            raise RuntimeError(
                f"Cannot generate report object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"report_id": report_id, "dataset_version_id": dataset_version_id},
        )

    def upload_file(
        self,
        object_name: str,
        path: Union[str, Path],
        connector_id: Optional[UUID] = None,
    ) -> tuple[requests.Response, bool, str]:
        """Upload a single file to the server.
        If file is bigger than 5Mb, it will send it by multipart.

        Arguments:
            path (str): Absolute path to the file
            object_name (str): Destination object name.
            connector_id (UUID): Connector on which you need to upload file, if it's not default connector.
        """
        if connector_id is None:
            connector_id = self.connector_id

        uploader = FileUploader(connector_id, self.session, self.host, self.headers)
        return uploader.upload(object_name, path)

    ##############################################################
    # ------------------------ DOWNLOAD ------------------------ #
    ##############################################################
    def init_download(
        self, object_name: str, connector_id: Optional[UUID] = None
    ) -> str:
        """Retrieve a presigned url of this object name in order to download it"""
        if connector_id is None:
            connector_id = self.connector_id

        payload = {"object_name": object_name}
        r = self.post(
            path=f"/api/object-storage/{connector_id}/retrieve_presigned_url",
            data=orjson.dumps(payload),
        )

        if r.status_code != 200:
            raise exceptions.DistantStorageError("Errors while getting a presigned url")

        r = r.json()
        if "presigned_url" not in r:
            raise exceptions.DistantStorageError(
                "Errors while getting a presigned url. Unparsable response"
            )

        return r["presigned_url"]

    def do_download_file(
        self,
        path: Union[str, Path],
        url: str,
        is_large: bool,
        force_replace: bool,
        retry_count: int = 1,
    ) -> bool:
        try:
            return self._do_download_file(path, url, is_large, force_replace)
        except (exceptions.NetworkError, ConnectionError) as e:
            # Here for retro compatibility
            raise exceptions.DownloadError(
                f"Could not download {url} into {path}"
            ) from e

    @retry((exceptions.NetworkError, ConnectionError))
    def _do_download_file(
        self,
        path: Union[str, Path],
        url: str,
        is_large: bool,
        force_replace: bool,
    ) -> bool:
        """Retrieve a presigned url of this object name in order to download it"""
        if os.path.exists(path) and not force_replace:
            return False

        parent_path = Path(path).parent.absolute()
        os.makedirs(parent_path, exist_ok=True)

        response = self.session.get(url, stream=is_large, timeout=DEFAULT_TIMEOUT)

        if response.status_code == 429 or (500 <= response.status_code < 600):
            raise exceptions.NetworkError(
                f"Response status code is {response.status_code}. Could not get {url}"
            )

        response.raise_for_status()

        total_length = response.headers.get("content-length")
        if total_length is None:
            raise exceptions.NetworkError(
                "Downloaded content is empty but response is 200"
            )

        with open(path, "wb") as handler:
            if not is_large:
                handler.write(response.content)
            else:
                for data in response.iter_content(chunk_size=4096):
                    handler.write(data)

        return True



import abc
import logging
from functools import partial
from typing import Generic, Optional, TypeVar

from orjson import orjson
from pydantic import BaseModel, Field

from picsellia import pxl_multithreading as mlt
from picsellia.compatibility import add_data_mandatory_query_parameters
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.utils import flatten

logger = logging.getLogger(__name__)


class BaseFilter(BaseModel):
    limit: Optional[int] = Field(default=None, gt=0)
    offset: Optional[int] = Field(default=None, ge=0)
    order_by: Optional[list[str]] = None


TItem = TypeVar("TItem", bound=Dao)
TFilter = TypeVar("TFilter", bound=BaseFilter)


class AbstractItemLister(abc.ABC, Generic[TItem, TFilter]):
    def __init__(self, connexion: Connexion):
        self.connexion = connexion

    def list_items(self, filters: TFilter) -> list[TItem]:
        if filters.has_list_of_items():
            return self._chunk_filter_and_fetch_items(filters)
        else:
            return self._list_items_with_pagination(filters)

    @abc.abstractmethod
    def _get_query_param_and_items_from_filters(
        self, filters: TFilter
    ) -> tuple[str, list]:
        pass

    @abc.abstractmethod
    def _get_path(self) -> str:
        pass

    @abc.abstractmethod
    def _build_item(self, data: dict) -> TItem:
        pass

    def _chunk_filter_and_fetch_items(self, filters: TFilter) -> list[TItem]:
        query_param, items = self._get_query_param_and_items_from_filters(filters)

        paginate_fn = partial(
            self._call_xget_for_one_chunk,
            filters.query,
            filters.order_by,
            query_param,
        )

        return flatten(
            mlt.do_chunk_called_function(
                items,
                f=paginate_fn,
                chunk_size=1000,
            )
        )

    def _call_xget_for_one_chunk(
        self,
        query: Optional[str],
        order_by: Optional[list[str]],
        query_param: str,
        items: list,
    ) -> list[TItem]:
        params = {}
        if order_by:
            params["order_by"] = order_by

        payload = {query_param: items}
        if query:
            payload["q"] = query

        return self._call_xget(params, payload)[0]

    def _call_xget(self, params: dict, payload: dict) -> tuple[list[TItem], int]:
        add_data_mandatory_query_parameters(payload)
        content = self.connexion.xget(
            self._get_path(),
            params=params,
            data=orjson.dumps(payload),
        ).json()
        return self._parse_response(content)

    def _list_items_with_pagination(self, filters: TFilter) -> list[TItem]:
        paginate_fn = partial(
            self._call_get_for_one_page, filters.query, filters.order_by
        )
        return mlt.do_paginate(filters.limit, filters.offset, None, paginate_fn)

    def _call_get_for_one_page(
        self,
        query: Optional[str],
        order_by: Optional[list[str]],
        limit: int,
        offset: int,
    ) -> tuple[list[TItem], int]:
        params = {"limit": limit, "offset": offset}
        if order_by:
            params["order_by"] = order_by
        if query:
            params["q"] = query
        return self._call_get(params)

    def _call_get(self, params: dict) -> tuple[list[TItem], int]:
        add_data_mandatory_query_parameters(params)
        content = self.connexion.get(self._get_path(), params=params).json()
        return self._parse_response(content)

    def _parse_response(self, content: dict) -> tuple[list[TItem], int]:
        return (
            [self._build_item(item) for item in content["items"]],
            content["count"],
        )
